# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems


class MlAssetsCreateFunctionInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self):
        """
        MlAssetsCreateFunctionInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'tags': 'list[TagRepository]',
            'name': 'str',
            'type': 'str',
            'description': 'str',
            'runtime_url': 'str',
            'sample_scoring_input': 'SampleScoringInput',
            'input_data_schema': 'InputDataSchema',
            'output_data_schema': 'OutputDataSchema'
        }

        self.attribute_map = {
            'tags': 'tags',
            'name': 'name',
            'type': 'type',
            'description': 'description',
            'runtime_url': 'runtime_url',
            'sample_scoring_input': 'sample_scoring_input',
            'input_data_schema': 'input_data_schema',
            'output_data_schema': 'output_data_schema'
        }

        self._tags = None
        self._name = None
        self._type = None
        self._description = None
        self._runtime_url = None
        self._sample_scoring_input = None
        self._input_data_schema = None
        self._output_data_schema = None

    @property
    def tags(self):
        """
        Gets the tags of this MlAssetsCreateFunctionInput.


        :return: The tags of this MlAssetsCreateFunctionInput.
        :rtype: list[TagRepository]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """
        Sets the tags of this MlAssetsCreateFunctionInput.


        :param tags: The tags of this MlAssetsCreateFunctionInput.
        :type: list[TagRepository]
        """
        self._tags = tags

    @property
    def name(self):
        """
        Gets the name of this MlAssetsCreateFunctionInput.


        :return: The name of this MlAssetsCreateFunctionInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this MlAssetsCreateFunctionInput.


        :param name: The name of this MlAssetsCreateFunctionInput.
        :type: str
        """
        self._name = name

    @property
    def type(self):
        """
        Gets the type of this MlAssetsCreateFunctionInput.
        Type of the function to be created. Only \"python\" type is supported as of now. If not specified, \"python\" is the default

        :return: The type of this MlAssetsCreateFunctionInput.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Sets the type of this MlAssetsCreateFunctionInput.
        Type of the function to be created. Only \"python\" type is supported as of now. If not specified, \"python\" is the default

        :param type: The type of this MlAssetsCreateFunctionInput.
        :type: str
        """
        self._type = type

    @property
    def description(self):
        """
        Gets the description of this MlAssetsCreateFunctionInput.


        :return: The description of this MlAssetsCreateFunctionInput.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this MlAssetsCreateFunctionInput.


        :param description: The description of this MlAssetsCreateFunctionInput.
        :type: str
        """
        self._description = description

    @property
    def runtime_url(self):
        """
        Gets the runtime_url of this MlAssetsCreateFunctionInput.


        :return: The runtime_url of this MlAssetsCreateFunctionInput.
        :rtype: str
        """
        return self._runtime_url

    @runtime_url.setter
    def runtime_url(self, runtime_url):
        """
        Sets the runtime_url of this MlAssetsCreateFunctionInput.


        :param runtime_url: The runtime_url of this MlAssetsCreateFunctionInput.
        :type: str
        """
        self._runtime_url = runtime_url

    @property
    def sample_scoring_input(self):
        """
        Gets the sample_scoring_input of this MlAssetsCreateFunctionInput.


        :return: The sample_scoring_input of this MlAssetsCreateFunctionInput.
        :rtype: SampleScoringInput
        """
        return self._sample_scoring_input

    @sample_scoring_input.setter
    def sample_scoring_input(self, sample_scoring_input):
        """
        Sets the sample_scoring_input of this MlAssetsCreateFunctionInput.


        :param sample_scoring_input: The sample_scoring_input of this MlAssetsCreateFunctionInput.
        :type: SampleScoringInput
        """
        self._sample_scoring_input = sample_scoring_input

    @property
    def input_data_schema(self):
        """
        Gets the input_data_schema of this MlAssetsCreateFunctionInput.


        :return: The input_data_schema of this MlAssetsCreateFunctionInput.
        :rtype: InputDataSchema
        """
        return self._input_data_schema

    @input_data_schema.setter
    def input_data_schema(self, input_data_schema):
        """
        Sets the input_data_schema of this MlAssetsCreateFunctionInput.


        :param input_data_schema: The input_data_schema of this MlAssetsCreateFunctionInput.
        :type: InputDataSchema
        """
        self._input_data_schema = input_data_schema

    @property
    def output_data_schema(self):
        """
        Gets the output_data_schema of this MlAssetsCreateFunctionInput.


        :return: The output_data_schema of this MlAssetsCreateFunctionInput.
        :rtype: OutputDataSchema
        """
        return self._output_data_schema

    @output_data_schema.setter
    def output_data_schema(self, output_data_schema):
        """
        Sets the output_data_schema of this MlAssetsCreateFunctionInput.


        :param output_data_schema: The output_data_schema of this MlAssetsCreateFunctionInput.
        :type: OutputDataSchema
        """
        self._output_data_schema = output_data_schema

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




"""Connection class which uses flask and request libraries to create a server
client combo
"""
#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging
import socket

import ibm_watson_machine_learning._wrappers.requests as requests
import asyncio
import signal
import random
import websockets
import pathlib
from websockets import exceptions
import string
import argparse
import threading
import time
import ssl
from _collections import deque
import sys
import traceback

from ibmfl.connection.connection import ConnectionStatus
from ibmfl.connection.connection import FLConnection, FLSender, FLReceiver
from ibmfl.message.message_type import MessageType
from ibmfl.message.message import Message
from ibmfl.message.serializer_types import SerializerTypes
from ibmfl.message.serializer_factory import SerializerFactory

logger = logging.getLogger(__name__)


# Maintained only on the party side
# Used to signal to websockets that there is a message ready to be sent to the aggregator
pSendEvt = threading.Event()

# Used to signal that there is a message that was received from the aggregator
pRecvEvt = threading.Event()



# Used to store messages to be sent to aggregator
party_sbuffer = deque([])
# Used to store messages received from aggregator
party_rbuffer = deque([])

pRouter = None


class WSConnection(FLConnection):

    def __init__(self, config):
        """
        Initializes the connection object
        :param config: dictionary of configuration provided to connection
        :type config: `dict`
        """

        pSendEvt.clear()
        pRecvEvt.clear()
        # comes from the config file
        self.started = False
        self.stopped = False
        self.settings = config
        self.receiver = None
        self.sender = None
        self.loop1 = None
        self.loop2 = None
        self.stop1 = None
        self.stop2 = None
        self.mainLoop = None
        self.flagForOutstandingMessageResponseToBeReceived = False
        # identity/name of the party

        if 'id' in config:
            self.party = config['id']
        else:
            # TODO: Fix error if no rts id
            self.party = 'DEFAULT'

        # Send Port is used for messages initiated from aggregator
        if 'wssendport' in config:
            self.wssendport = config['wssendport']
        else:
            # set a default value to be aggregator
            self.wssendport = "/aggregator"
        # Recv port is used for messages initiated from party
        if 'wsrecvport' in config:
            self.wsrecvport = config['wsrecvport']
        else:
            # set a default value to be aggregator
            self.wsrecvport = "/aggregator1"


    async def PartySendLoop(self):
        """
        This is started at the party.
        It reads from the party_sbuffer, and synchronously sends each message to the aggregator
        and waits for a response from the aggregator.
        SendLoop is a misnomer, like in Flask Connection.
        It merely means that messages are initiated by the party.
        """

        uri = "wss://" + self.aggInfo['ip'] + str(self.wsrecvport)
        logger.debug("PartySendLoop started = " + uri)
        ssl_context_party = True
        if self.ssl_context is not None:
            ssl_context_party = ssl.SSLContext()

        while not self.stopped:
            if len(party_sbuffer) == 0:
                logger.info("PartySendLoop: Holding for message to send" )
                pSendEvt.wait()
            try:
                logger.info("PartySendLoop: Number of active messages ready to send: " + str(len(party_sbuffer)))
                msg = party_sbuffer.popleft()
            except IndexError:
                logger.debug("No message to send")
                continue
            pSendEvt.clear()
            logger.debug("PartySendLoop: Sending message to aggregator")
            headers = None
            if self.authToken != None:
                headers = { 'Authorization': self.authToken, 'rtsid': self.party }
            try:
                async with websockets.connect(uri, max_size=2 ** 29, read_limit=2 ** 29, write_limit=2**29, ssl=ssl_context_party, extra_headers=headers) as websocket:

                    if self.flagForOutstandingMessageResponseToBeReceived == False:
                        logger.debug(
                            "PartySendLoop: Attempting to send message to aggregator")
                        try:
                            await websocket.send(msg)

                        except Exception as ex:
                            logger.error("PartySendLoop exception when attempting to send message: " + str(ex))
                            party_sbuffer.appendleft(msg)
                            raise
                        logger.debug(
                            "PartySendLoop: Message sent successfully; awaiting message response")

                    self.flagForOutstandingMessageResponseToBeReceived = True
                    resp = await websocket.recv()
                    self.flagForOutstandingMessageResponseToBeReceived = False
                    party_rbuffer.append(resp)
                    pRecvEvt.set()

            except websockets.exceptions.ConnectionClosedError as ex:
                if self.stopped != True:
                    logger.error(
                        "PartySendLoop : Connection closed abnormally. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.ConnectionClosedOK as ex:
                if self.stopped != True:
                    logger.error(
                        "PartySendLoop : Connection closed unexpectedly. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.WebSocketException as ex:
                logger.error(
                    "PartySendLoop : WebsocketException. Details: " + str(ex))
                return False
            except socket.gaierror as ex:
                logger.error("PartySendLoop : Error connecting to " + uri + ". Exception is " + str(ex))
                return False
            except Exception as ex:
                logger.error("PartySendLoop : Exception is " + str(ex))
                return False
        logger.debug("PartySendLoop ended ")
        return True

    async def PartyRecvLoop(self):
        """
        This is started at the party.
        It establishes a persistent connection to the aggregator by sending the name of the party
        The aggregator can then send messages to the party synchronously
        to which the party replies
        """
        pName = self.party
        uri = "wss://" + self.aggInfo['ip'] + str(self.wssendport)
        logger.debug("PartyRecvLoop started = " + uri)

        ssl_context_party2 = True
        if self.ssl_context is not None:
            ssl_context_party2 = ssl.SSLContext()

        headers = None
        if self.authToken != None:
            headers = { 'Authorization': self.authToken }
        while not self.stopped:
            try:
                async with websockets.connect(uri, close_timeout=100, max_size=2 ** 29, read_limit=2 ** 29, write_limit=2**29, ssl=ssl_context_party2, extra_headers=headers) as websocket:
                    await websocket.send(pName)
                    async for rmess in websocket:
                        if rmess == "HEARTBEAT":
                            logger.info("Received Heartbeat from Aggregator")
                            continue
                        serializer = SerializerFactory(
                            SerializerTypes.JSON_PICKLE).build()
                        recv_msg = serializer.deserialize(rmess)
                        logger.debug(
                            "PartyRecvLoop : Received " + str(recv_msg))
                        request_path = str(recv_msg.message_type)
                        logger.debug(
                            "REQUEST PATH IN PARTY IS " + str(request_path))

                        # check if error
                        if recv_msg.message_type == MessageType.ERROR_AUTH.value :
                            logger.info(
                            "PartyRecvLoop : Auth Error. Please verify training ID, RTS, project permission and token")
                            self.stop()
                            break;

                        handler, kwargs = self.router.get_handler(
                            request_path=request_path)

                        if handler is None:
                            logger.error(
                                'PartyRecvLoop : Invalid Request ! Routing it to default handler')
                            handler, kwargs = self.router.get_handler(
                                request_path='default')
                        try:
                            res_message = handler(recv_msg)
                        except Exception as ex:
                            res_message = Message()
                            data = {'status': 'error', 'message': str(ex)}
                            res_message.set_data(data)

                        response = serializer.serialize(res_message)
                        await websocket.send(response)
                        logger.info("PartyRecvLoop : sent response")

                        if recv_msg.message_type == MessageType.STOP.value :
                            logger.info(
                            "PartyRecvLoop : Received STOP message" )
                            self.stop()
                            break;

            except websockets.exceptions.ConnectionClosedError as ex:
                if self.stopped != True:
                    logger.error(
                        "PartyRecvLoop : Connection closed abnormally. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.ConnectionClosedOK as ex:
                if self.stopped != True:
                    logger.error(
                        "PartyRecvLoop : Connection closed unexpectedly. Exception is " + str(ex))
                    logger.error("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.WebSocketException as ex:
                logger.error(
                    "PartyRecvLoop : WebsocketException. Details: " + str(ex))
                return False
            except socket.gaierror as ex:
                logger.error("PartyRecvLoop : Error connecting to " + uri + ". Exception is " + str(ex))
                return False
            except Exception as ex:
                logger.error("PartyRecvLoop : Exception is " + str(ex))
                return False
        logger.debug("PartyRecvLoop ended ")
        return True

    def PartySendLoopThread(self, loop):
        logger.info("**** PartySendLoopThread")
        # Need to host the loop in its own thread for asyncio to work properly
        pslt_result = loop.run_until_complete(self.PartySendLoop())
        logger.debug("**** PartySendLoopThread RESULT = " + str(pslt_result))
        if pslt_result == False:
            #if error when sending message, then unblock mainThread so it can fail.
            pRecvEvt.set()
            self.stop()

    def PartyRecvLoopThread(self, loop):
        logger.info("**** PartyRecvLoopThread")
        # Need to host the loop in its own thread for asyncio to work properly
        prlt_result = loop.run_until_complete(self.PartyRecvLoop())
        logger.debug("**** PartyRecvLoopThread RESULT = " + str(prlt_result))
        if prlt_result == False:
            self.stop()


    def initialize(self, router, aggInfo=None, ssl_context=None, authToken=None):
        """Initialize receiver and sender """
        self.router = router
        self.ssl_context = ssl_context

        if aggInfo == None:
            logger.error("Aggregator information not specified. Exiting.")
            exit()
        logger.info("WSConnection : Initialize Party Communications")
        self.party = self.settings['id']
        self.aggInfo = aggInfo
        self.authToken = authToken

        # Each websocket connection needs a separate asyncio event loop
        self.loop1 = asyncio.new_event_loop()
        t1 = threading.Thread(
            target=self.PartySendLoopThread, args=(self.loop1,), daemon=True)
        t1.start()

        # Each websocket connection needs a separate asyncio event loop
        self.loop2 = asyncio.new_event_loop()
        t2 = threading.Thread(
            target=self.PartyRecvLoopThread, args=(self.loop2,), daemon=True)
        t2.start()

    def initialize_receiver(self, router=None):
        """Basically does nothing
        :param router: Router object describing the routes for each request
            which are passed down to PH
        :type router: `Router`
        """
        # print(router)
        self.router = router
        pRouter = router
        self.receiver = WSReceiver(router, self.settings, self)
        self.receiver.connection = self
        logger.debug('Receiver Initialized')
        self.receiver.initialize()
        self.status = ConnectionStatus.INITIALIZED

    def initialize_sender(self):
        """BAsically does nothing
        """
        self.sender = WSSender(self.settings)
        self.sender.initialize()
        self.status = ConnectionStatus.INITIALIZED

    def start(self):
        """
        Set self.started to True
        """
        # self.start_receiver()
        self.started = True
        self.status = ConnectionStatus.STARTED
        self.SENDER_STATUS = ConnectionStatus.STARTED

    def stop(self):
        """
        set self.stopped to true
        """
        self.stopped = True
        logger.debug('Stopping Receiver and Sender')
        pSendEvt.set()
        self.status = ConnectionStatus.STOPPED
        try:
            if self.stop1 is not None:
                self.stop1.cancel()
            if self.stop2 is not None:
                self.stop2.set_result(None)
            if self.mainLoop is not None:
                self.mainLoop.call_soon_threadsafe(self.mainLoop.stop)
        except Exception as ex:
            logger.error("Exception when stopping connection: " + str(ex))

    def get_connection_config(self):
        pass

class WSReceiver(FLReceiver):

    def __init__(self, router, settings, connection):
        """
        Does nothing
        """

        self.router = router
        self.settings = settings
        self.connection = connection

    def shutdown_server(self):
        pass

    def initialize(self):
        """
        Does nothing
        """
        logger.debug('Initializing WSReceiver')


    def receive_message(self, party):
        logger.error('OOOOO We are in trouble. This should not be called')

    def start(self):
        pass

    def stop(self):
        pass


class WSSender(FLSender):

    """
    Basic implementation using the request package of python to send
    requests to Rest endpoints.
    """

    def __init__(self, settings):
        """
        Does nothing
        """
        self.settings = settings
        logger.debug("WS Sender Init. My info is " + str(settings))

    def initialize(self):
        """Does all the setups required for the sender. This method should also
        check for availability of resources, open ports, certificates etc.,
        if required
        """
        logger.info('Websockets Sender initialized')

    def send_message(self, destination, message):
        """
        used for sending all the requests. Message object should be
        validated and endpoint should be decided based on message codes.
        :param destination: information about the destination to which message
        should be forwarded
        :type destination: `dict`
        :param message: message object constructed by aggregator/party
        :type message: `Message`
        :return : response object
        :rtype : `Response`
        """
        if 'id' in self.settings:
            message.add_sender_info(self.settings['id'])
        else:
            # set a default value to be Aggregator
            message.add_sender_info('Aggregator')

        serializer = SerializerFactory(SerializerTypes.JSON_PICKLE).build()
        message = serializer.serialize(message)

        logger.info("Sending serialized message to aggregator")
        party_sbuffer.append(message)
        pSendEvt.set()
        pRecvEvt.wait()
        response = None
        try:
            response = party_rbuffer.popleft()
        except IndexError:
            logger.debug("[WSSender] No message response received, after attempt to send message")

        pRecvEvt.clear()
        response_message = None
        if response is not None:
            # inspect response
            if isinstance(response, str):
                logger.info('Received String message as response: ' + response)
                response_message = Message(
                    MessageType.REGISTER.value, data={'status': 'success', 'response': response})
            else:
                # deserialize response
                response_message = serializer.deserialize(response)
                logger.info(
                    'Received serialized message as response: ' + str(response_message))

        return response_message

    def cleanup(self):
        """
        Cleanup the Sender object and close the hanging connections if any
        """
        logger.info('Cleaning up Websockets Client')



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging
from abc import ABCMeta, abstractmethod

from pyspark import SparkContext
from pyspark.ml.common import _py2java
from pyspark.ml.common import inherit_doc
from pyspark.ml.pipeline import Transformer, Estimator
from pyspark.ml.util import JavaMLWritable, MLReadable
from pyspark.ml.wrapper import JavaTransformer, JavaEstimator, JavaWrapper, JavaParams
from pyspark.ml.wrapper import _jvm

from mlpipelinepy.edge import DataEdge, MetaEdge
from mlpipelinepy.exception import UnsupportedNodeTypeError
from mlpipelinepy.javaproxy import TransformerProxyML, EstimatorProxyML
from mlpipelinepy.serialization import PipelineJavaMLReader

__all__ = ['MLPipeline', "MLPipelineModel", "SparkDataSources"]

logger = logging.getLogger("mlpipelinepy")


@inherit_doc
class MLPipelineBasic(JavaWrapper):
    """
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        super(MLPipelineBasic, self).__init__()

    def bind(self, from_node_id, to_node_id, from_edge_id=None):
        """
         Binds 'to_node_id' node with 'from_edge_id' edge of the 'from_node_id' node
        :param from_node_id: the source node
        :param to_node_id: the destination node
        :param from_edge_id: the edge of the source node
        :returns: pipeline instance
        """
        if from_edge_id is None:
            self._java_obj = self._java_obj.bind(from_node_id, to_node_id)
        else:
            self._java_obj = self._java_obj.bind(from_node_id, from_edge_id, to_node_id)
        return self

    def transform(self, spark_data_source, extra=None):
        """
        Transforms the input dataset with optional parameters.
        :param spark_data_source: SparkDataSouce object containing mapping between nodeid and dataframe
        :param extra: an optional param map that overrides embedded params.
        :returns: transformed dataset
        """
        try:
            jvm_object = _jvm()
            java_p_map = self._to_scala_param_map(extra)
            seq_m = self._java_obj.transform(spark_data_source.to_java_object(), java_p_map)
            _jDataEdgeList = jvm_object.com.ibm.analytics.wml.pipeline.pythonbinding.Helper.scalaToJavaList(seq_m)
            pDataEdgeList = []
            for _sDataEdge in _jDataEdgeList:
                pDataEdgeList.append(DataEdge(_sDataEdge))
            return pDataEdgeList
        except Exception as e:
            logger.exception(e)
            raise

    @abstractmethod
    def copy(self, extra=None):
        raise NotImplementedError()

    def node_ids(self):
        jvm_object = _jvm()
        s_nodsIds = self._java_obj.nodeIds()
        j_list = jvm_object.com.ibm.analytics.wml.pipeline.pythonbinding.Helper.scalaToJavaList(s_nodsIds)
        return map(str, j_list)

    def _to_scala_param_map(self, extra=None):
        sc = SparkContext._active_spark_context
        jvmObject = _jvm()
        p_map = jvmObject.org.apache.spark.ml.param.ParamMap.empty()
        if extra is None:
            extra = dict()
        if isinstance(extra, dict):
            for param in extra:
                try:
                    uid = param.parent
                    name = param.name
                    java_param_try = self._java_obj.nodeParam(uid, name)
                    if java_param_try.isSuccess():
                        java_param = java_param_try.get()
                        value = extra[param]
                        java_value = _py2java(sc, value)
                        p_map.put([java_param.w(java_value)])
                except Exception as e:
                    logger.exception(e)
                    raise
        else:
            raise ValueError("Params must be a param map, but got %s." % type(extra))
        return p_map

    def schema(self, spark_data_source, extra=None):
        """Returns the list of MetaEdge of this :class:`MLPipelineBasic` `.
        >>> pipeline.schema(SparkDataSources({"id": df}))
        MetaEdge (is, schema)
        """
        try:
            java_p_map = self._to_scala_param_map(extra)
            seq_meta_struc_type = self._java_obj.schema(spark_data_source.to_java_object(), java_p_map)
            jvm_object = _jvm()
            j_list = jvm_object.com.ibm.analytics.wml.pipeline.pythonbinding.Helper.scalaToJavaList(seq_meta_struc_type)
            return [MetaEdge(item) for item in j_list]
        except AttributeError as e:
            raise Exception(
                "Unable to parse datatype from schema. %s" % e)


@inherit_doc
class MLPipeline(MLPipelineBasic, JavaParams, JavaMLWritable, MLReadable):
    """
    MLPipeline wrapper class
    """

    def __init__(self, nodes=[]):
        super(MLPipeline, self).__init__()
        jvm_obj = _jvm()
        java_nodes = self._nodes_to_java_nodes(nodes, jvm_obj)
        _sNodesList = jvm_obj.com.ibm.analytics.wml.pipeline.pythonbinding.Helper.javaToScalaList(java_nodes)
        self._java_obj = jvm_obj.com.ibm.analytics.wml.pipeline.spark.MLPipeline.apply(_sNodesList)

    @classmethod
    def _to_java_node(cls, node, jvm_obj):
        """
        :param node: python estimator / transformer node
        :param jvm_obj: JVM object
        :return: java computation node required by scala MLPipeline instance
        """
        java_obj = None
        if isinstance(node, Transformer):
            if isinstance(node, JavaTransformer):
                node._transfer_params_to_java()
                java_obj = node._java_obj
            else:
                new_transformer_wrapper = TransformerProxyML(node)
                java_obj = new_transformer_wrapper.to_java_object()
            return jvm_obj.com.ibm.analytics.wml.pipeline.spark.package.transformer2PipelineNode(java_obj)
        elif isinstance(node, Estimator):
            if isinstance(node, JavaEstimator):
                node._transfer_params_to_java()
                java_obj = node._java_obj
            else:
                new_estimator_wrapper = EstimatorProxyML(node)
                java_obj = new_estimator_wrapper.to_java_object()
            return jvm_obj.com.ibm.analytics.wml.pipeline.spark.package.estimator(java_obj)
        else:
            raise UnsupportedNodeTypeError("Unsupport node type %s" % node.__class__.__name__)

    def _nodes_to_java_nodes(self, nodes, jvm_object):
        jNodes = []
        for node in nodes:
            jNodes.append(self._to_java_node(node, jvm_object))
        return jNodes

    def copy(self, extra=None):
        """
        Creates a copy of this instance.

        :param extra: extra parameters
        :returns: new instance
        """
        java_p_map = self._to_scala_param_map(extra)
        new_java_obj = self._java_obj.copy(java_p_map)
        ret_val = MLPipeline([])
        ret_val._java_obj = new_java_obj
        return ret_val

    def fit(self, spark_data_source, extra=None):
        """
        Fits a pipeline with optional parameters and produce a pipeline model

        :param params: an optional param map that overrides embedded params.
        :returns: fitted pipeline model
        """
        if extra is None:
            extra = dict()
        java_p_map = self._to_scala_param_map(extra)

        if isinstance(extra, dict):
            try:
                return MLPipelineModel(self, self._java_obj.fit(spark_data_source.to_java_object(), java_p_map))
            except Exception as e:
                logger.exception(e)
                raise
        else:
            raise ValueError("Params must be a param map, but got %s." % type(extra))

    def _to_java(self):
        return self._java_obj

    @classmethod
    def _from_java(cls, java_obj):
        """
        Given a Java Pipeline, create and return a Python wrapper of it.
        Used for ML persistence.
        """
        # Create a new instance of this stage.
        py_obj = cls()
        py_obj._java_obj = java_obj
        return py_obj

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return PipelineJavaMLReader(cls, "com.ibm.analytics.wml.pipeline.spark.MLPipeline")


class MLPipelineModel(MLPipelineBasic, JavaMLWritable, MLReadable):
    """
    PipelineModel wrapper class
    """

    def __init__(self, parent_pipeline=None, java_model=None):
        super(MLPipelineModel, self).__init__()
        self._java_obj = java_model
        self._parent = parent_pipeline

    def copy(self, extra=None):
        java_p_map = self._to_scala_param_map(extra)
        new_java_obj = self._java_obj.copy(java_p_map)
        return MLPipelineModel(new_java_obj)

    @property
    def parent(self):
        return self._parent

    def _to_java(self):
        return self._java_obj

    @classmethod
    def _from_java(cls, java_obj):
        """
        Given a Java Pipeline, create and return a Python wrapper of it.
        Used for ML persistence.
        """
        # Create a new instance of this stage.
        py_obj = cls()
        py_obj._java_obj = java_obj
        if java_obj is None and java_obj.parentPipeline().isDefined():
            py_parent = MLPipeline()
            py_parent._java_obj = java_obj.parentPipeline().get()
            py_obj._parent = py_parent
        return py_obj

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return PipelineJavaMLReader(cls, "com.ibm.analytics.wml.pipeline.spark.MLPipelineModel")


class SparkDataSources(object):
    """
    SparkDataSources wrapper class
    """

    def __init__(self, idDfMapping=dict()):
        super(SparkDataSources, self).__init__()
        self._idDfMapping = idDfMapping

    def to_java_object(self):
        jvm_object = _jvm()
        elems = []
        for key, value in list(self._idDfMapping.items()):
            elems.append(jvm_object.scala.Tuple2(key, value._jdf))

        scalaList = jvm_object.com.ibm.analytics.wml.pipeline.pythonbinding.Helper.javaToScalaList(elems)
        return jvm_object.com.ibm.analytics.wml.pipeline.spark.SparkDataSources(
            jvm_object.com.ibm.analytics.wml.pipeline.pythonbinding.Helper.scalaListOfPairsToMap(scalaList))



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function

from sklearn._loss.loss import (
    _LOSSES,
    AbsoluteError,
    BaseLoss,
    HalfBinomialLoss,
    HalfMultinomialLoss,
    HalfPoissonLoss,
    HalfSquaredError,
    PinballLoss,
)


def is_classifier(cls):
    """
    Auxillary function to validate whether a given object is a
    classification-based model. (Can be used for the Fusion Handler as well as
    the Local Training Handler).

    :param cls: Object of interest to identify the type of model.
    :type  cls: `Object`
    :return `bool`
    """
    return cls.loss == "auto" or cls.loss == "binary_crossentropy" or cls.loss == "categorical_crossentropy"


_LOSSES = _LOSSES.copy()
_LOSSES.update(
    {
        "least_squares": HalfSquaredError,
        "least_absolute_deviation": AbsoluteError,
        "binary_crossentropy": HalfBinomialLoss,
        "categorical_crossentropy": HalfMultinomialLoss,
    }
)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging
from ibmfl.party.training.local_training_handler import \
    LocalTrainingHandler

logger = logging.getLogger(__name__)


class FedAvgLocalTrainingHandler(LocalTrainingHandler):

    def train(self,  fit_params=None):
        """
        Train locally using fl_model. At the end of training, a
        model_update with the new model information is generated and
        send through the connection.

        :param fit_params: (optional) Query instruction from aggregator
        :type fit_params: `dict`
        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        train_data, (_) = self.data_handler.get_data()
        _train_count = train_data[0].shape[0]

        self.update_model(fit_params['model_update'])

        self.get_train_metrics_pre()

        logger.info('Local training started...')

        self.fl_model.fit_model(train_data, fit_params, local_params=self.hyperparams)

        update = self.fl_model.get_model_update()
        update.add('train_counts', _train_count)

        logger.info('Local training done, generating model update...')

        self.get_train_metrics_post()

        return update



#  (C) Copyright IBM Corp. 2023.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure

_DEFAULT_LIST_LENGTH = 50


class TaskCredentials(WMLResource):
    """Store and manage your task credentials."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)

    def get_details(self, task_credentials_uid=None, project_id=None, space_id=None):
        """Get task credentials details. If no task_credentials_uid is passed, details for all task credentials
        will be returned.

        :param task_credentials_uid: UID of task credentials to be fetched
        :type task_credentials_uid: str, optional

        :param project_id: UID of project to be used for filtering
        :type project_id: str, optional

        :param space_id: UID of space to be used for filtering
        :type space_id: str, optional

        :return: created task credentials details
        :rtype: dict (if task_credentials_uid is not None) or {"resources": [dict]} (if task_credentials_uid is None)

        **Example**

        .. code-block:: python

            task_credentials_details = client.task_credentials.get_details(task_credentials_uid)

        """
        if self._ICP:
            raise WMLClientError(u'Task Credentials API is supported on Cloud only.')

        # TaskCredentials._validate_type(task_credentials_uid, u'task_credentials_uid', STR_TYPE, False)

        if task_credentials_uid:
            response = requests.get(
                self._client.service_instance._href_definitions.get_task_credentials_href(task_credentials_uid),
                headers=self._client._get_headers())

            return self._handle_response(200, u'get task credentials details', response)
        else:
            params = {}

            if project_id:
                params['scope.project_id'] = project_id
            elif space_id:
                params['scope.space_id'] = space_id

            response = requests.get(self._client.service_instance._href_definitions.get_task_credentials_all_href(),
                                    params=params,
                                    headers=self._client._get_headers())

            return {'resources': self._handle_response(200, u'get task credentials details',
                                                       response).get('credentials', {})}

    def store(self, project_id=None, space_id=None):
        """Store current credentials using Task Credentials API to use with long run tasks. Supported only on Cloud.

        :param project_id: UID of project which become a scope for saved credentials
        :type project_id: str, optional

        :param space_id: UID of space which become a scope for saved credentials
        :type space_id: str, optional

        :return: metadata of the stored task credentials
        :rtype: dict

        **Example**

        .. code-block:: python

            task_credentials_details = client.task_credentials.store()

        """
        if self._ICP:
            raise WMLClientError(u'Task Credentials API is supported on Cloud only.')

        href = self._client.service_instance._href_definitions.get_task_credentials_all_href()

        scope = {}

        if project_id:
            scope['project_id'] = project_id
        elif space_id:
            scope['space_id'] = space_id

        creation_response = requests.post(href,
                                          params=self._client._params(),
                                          headers=self._client._get_headers(),
                                          json={"name": "Python API generated task credentials",
                                                "description": "Python API generated task credentials.",
                                                "type": "iam_api_key",
                                                "scope": scope,
                                                "secret": {
                                                    "api_key": self._client.wml_credentials['apikey']
                                                }})

        return self._handle_response(201, u'creating task credentials', creation_response)

    def list(self, limit=None, project_id=None, space_id=None, return_as_df=True):
        """Print task credentials in table format.

        :param limit: limit number of fetched records
        :type limit: int, optional

        :param project_id: UID of project to be used for filtering
        :type project_id: str, optional

        :param space_id: UID of space to be used for filtering
        :type space_id: str, optional

        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed assets or None if return_as_df is False
        :rtype: pandas.DataFrame or None


        **Example**

        .. code-block:: python

            client.task_credentials.list()

        """
        if self._ICP:
            raise WMLClientError(u'Task Credentials API is supported on Cloud only.')

        details = self.get_details(project_id=project_id, space_id=space_id)

        task_credentials_details = details["resources"]
        task_credentials_values = [
            (m[u'name'], m[u'id'], m[u'scope']) for
            m in task_credentials_details]

        table = self._list(task_credentials_values, [u'NAME', u'ASSET_ID', u'TYPE'], limit, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    @staticmethod
    def get_id(task_credentials_details):
        """Get Unique Id of task credentials.

        :param task_credentials_details: metadata of the task credentials
        :type task_credentials_details: dict

        :return: Unique Id of task credentials
        :rtype: str

        **Example**

        .. code-block:: python

            task_credentials_uid = client.task_credentials.get_id(task_credentials_details)

        """
        return task_credentials_details['id']

    def delete(self, task_credentials_uid):
        """Delete a software specification.

        :param task_credentials_uid: Unique Id of task credentials
        :type task_credentials_uid: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.task_credentials.delete(task_credentials_uid)

        """
        if self._ICP:
            raise WMLClientError(u'Task Credentials API is supported on Cloud only.')

        TaskCredentials._validate_type(task_credentials_uid, u'task_credentials_uid', str, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_task_credentials_href(task_credentials_uid),
                                   params=self._client._params(),
                                   headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete task credentials', response)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import PKG_EXTN_DETAILS_TYPE
from ibm_watson_machine_learning.metanames import PkgExtnMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os,json

_DEFAULT_LIST_LENGTH = 50


class PkgExtn(WMLResource):
    """Store and manage software Packages Extension specs."""

    ConfigurationMetaNames = PkgExtnMetaNames()
    """MetaNames for Package Extensions creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    def get_details(self, pkg_extn_id):
        """Get package extensions details.

        :param pkg_extn_id: Unique Id of package extension
        :type pkg_extn_id: str

        :return: details of the package extensions
        :rtype: dict

        **Example**

        .. code-block:: python

            pkg_extn_details = client.pkg_extn.get_details(pkg_extn_id)

        """
        PkgExtn._validate_type(pkg_extn_id, u'pkg_extn_id', str, True)

        response = requests.get(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_id),
                                params=self._client._params(),
                                headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(self._handle_response(200, u'get hw spec details', response))
        else:
            return self._handle_response(200, u'get hw spec details', response)

    def store(self, meta_props, file_path):
        """Create a package extensions.

        :param meta_props: meta data of the package extension. To see available meta names use:

            .. code-block:: python

                client.package_extensions.ConfigurationMetaNames.get()

        :type meta_props: dict

        :param file_path:  path to file which will be uploaded as package extension
        :type file_path: str

        :return: metadata of the package extensions
        :rtype: dict

        **Example**

        .. code-block:: python

            meta_props = {
                client.package_extensions.ConfigurationMetaNames.NAME: "skl_pipeline_heart_problem_prediction",
                client.package_extensions.ConfigurationMetaNames.DESCRIPTION: "description scikit-learn_0.20",
                client.package_extensions.ConfigurationMetaNames.TYPE: "conda_yml"
            }

            pkg_extn_details = client.package_extensions.store(meta_props=meta_props, file_path="/path/to/file")

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        PkgExtn._validate_type(meta_props, u'meta_props', dict, True)
        pkg_extn_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        pkg_extn_meta_json = json.dumps(pkg_extn_meta)

        PkgExtn._validate_type(file_path, u'file_path', str, True)
        #Step1  : Create an asset
        print("Creating package extensions")
        href = self._client.service_instance._href_definitions.get_pkg_extns_href()

        creation_response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), data=pkg_extn_meta_json)

        pkg_extn_details = self._handle_response(201, u'creating new package_extensions', creation_response)

        # Step2: upload pkg extension file to presigned url
        if creation_response.status_code == 201:
            pkg_extn_asset_id = pkg_extn_details["metadata"]["asset_id"]
            pkg_extn_presigned_url = pkg_extn_details["entity"]["package_extension"]["href"]

            FILE_SIZE_LIMIT = int(0.3 * 1073741824)

            def read_in_chunks(chunk_size=FILE_SIZE_LIMIT):
                with open(file_path, 'rb') as file_object:
                    while True:
                        data = file_object.read(chunk_size)
                        if not data:
                            break
                        yield data

            def send_multipart_request(file, url):
                content_path = os.path.abspath(file)
                content_size = os.stat(content_path).st_size

                index = 0
                headers = {}

                try:
                    for chunk in read_in_chunks(FILE_SIZE_LIMIT):
                        offset = index + len(chunk)
                        headers['Content-Range'] = 'bytes %s-%s/%s' % (index, offset - 1, content_size)
                        index = offset

                        response = requests.put(
                            url,
                            files={"file": (file, chunk, 'application/octet-stream')}
                        )
                except Exception as e:
                    deletion_response = requests.delete(
                        self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_asset_id),
                        params=self._client._params(),
                        headers=self._client._get_headers()
                    )
                    print(deletion_response.status_code)
                    raise WMLClientError("Failed while reading a file.", e)

                return response

            # chunking, when finally it will work on service side. to confirm it's working, the downloaded file needs to be the same as 400MB+ upladed file
            #
            # put_response = send_multipart_request(file_path,
            #   (self._client.service_instance._wml_credentials['url'] if self._client.ICP else "") + pkg_extn_presigned_url)

            href = (self._client.service_instance._wml_credentials['url'] if self._client.ICP else "") + pkg_extn_presigned_url

            # if pkg_extn_details["entity"]["package_extension"]["type"] in ["conda_yml", "pip_zip"]:
            try:
                if os.stat(file_path).st_size == 0:
                    raise WMLClientError('Package extension file cannot be empty')

                with open(file_path, 'rb') as file_object:
                    if not self._client.ICP:
                        put_response = requests.put(
                            href,
                            data=file_object.read()
                        )
                    else:
                        put_response = requests.put(
                            href,
                            files={'file': (file_path, file_object.read(), 'application/octet-stream')}
                        )
            except Exception as e:
                deletion_response = requests.delete(
                    self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_asset_id),
                    params=self._client._params(),
                    headers=self._client._get_headers()
                )
                print(deletion_response.status_code)
                raise WMLClientError("Failed while reading a file.", e)

            if put_response.status_code == 201 or put_response.status_code == 200:
                # Step3: Mark the upload complete
                complete_response = requests.post(
                    self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_asset_id) + "/upload_complete",
                    headers=self._client._get_headers(),
                    params=self._client._params()
                )
                if complete_response.status_code == 204:
                    print("SUCCESS")
                    return self._get_required_element_from_response(pkg_extn_details)
                else:
                    #print(complete_response.text) # remove print later
                    self._delete(pkg_extn_asset_id)
                    raise WMLClientError("Failed while creating a package extensions " + complete_response.text)
            else:
                self._delete(pkg_extn_asset_id)
                raise WMLClientError("Failed while creating a package extensions " + put_response.text)
        else:
            raise WMLClientError("Failed while creating a package extensions " + creation_response.text)

    def list(self, return_as_df=True):
        """List package extensions in a table format.

        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed package extensions or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        .. code-block:: python

            client.package_extensions.list()

        """

        href = self._client.service_instance._href_definitions.get_pkg_extns_href()

        response = requests.get(href, params=self._client._params(), headers=self._client._get_headers())

        self._handle_response(200, u'list pkg_extn', response)
        asset_details = self._handle_response(200, u'list assets', response)["resources"]
        pkg_extn_values = [
            (m[u'metadata'][u'name'],
             m[u'metadata'][u'asset_id'],
             m[u'entity'][u'package_extension'][u'type'],
             m[u'metadata'][u'created_at']) for
            m in asset_details]

        table = self._list(pkg_extn_values, [u'NAME', u'ASSET_ID', u'TYPE', u'CREATED_AT'], None, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    @staticmethod
    def get_uid(pkg_extn_details):
        """Get Unique Id of package extensions.

        *Deprecated:* Use ``get_id(pkg_extn_details)`` instead.

        :param pkg_extn_details: details of the package extensions
        :type pkg_extn_details: dict

        :return: Unique Id of package extension
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.package_extensions.get_uid(pkg_extn_details)

        """
        PkgExtn._validate_type(pkg_extn_details, u'pkg_extn_details', object, True)
        #print(pkg_extn_details)
        PkgExtn._validate_type_of_details(pkg_extn_details, PKG_EXTN_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(pkg_extn_details, u'pkg_extn_details',
                                                           [u'metadata', u'asset_id'])

    @staticmethod
    def get_id(pkg_extn_details):
        """Get Unique Id of package extensions.

        :param pkg_extn_details: details of the package extensions
        :type pkg_extn_details: dict

        :return: Unique Id of package extension
        :rtype: str

        **Example**

        .. code-block:: python

            asset_id = client.package_extensions.get_id(pkg_extn_details)

        """

        return PkgExtn.get_uid(pkg_extn_details)

    def get_uid_by_name(self, pkg_extn_name):
        """Get UID of package extensions.

        *Deprecated:* Use ``get_id_by_name(pkg_extn_name)`` instead.

        :param pkg_extn_name: name of the package extension
        :type pkg_extn_name: str

        :return: Unique Id of package extension
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.package_extensions.get_uid_by_name(pkg_extn_name)

        """
        PkgExtn._validate_type(pkg_extn_name, u'pkg_extn_name', str, True)


        parameters = self._client._params()
        parameters.update(name=pkg_extn_name)

        response = requests.get(self._client.service_instance._href_definitions.get_pkg_extns_href(),
                                params=parameters,
                                headers=self._client._get_headers())

        if response.status_code == 200:
            total_values = self._handle_response(200, u'get pkg extn', response)["total_results"]
            if total_values != 0:
                pkg_extn_details = self._handle_response(200, u'get pkg extn', response)["resources"]
                return pkg_extn_details[0][u'metadata'][u'asset_id']
            else:
                return "Not Found"

    def get_id_by_name(self, pkg_extn_name):
        """Get ID of package extensions.

        :param pkg_extn_name:  name of the package extension
        :type pkg_extn_name: str

        :return: Unique Id of package extension
        :rtype: str

        **Example**

        .. code-block:: python

            asset_id = client.package_extensions.get_id_by_name(pkg_extn_name)

        """
        return PkgExtn.get_uid_by_name(self, pkg_extn_name)


    @staticmethod
    def get_href(pkg_extn_details):
        """Get url of stored package extensions.

        :param pkg_extn_details: details of the package extensions
        :type pkg_extn_details: dict

        :return: href of package extension
        :rtype: str

        **Example**

        .. code-block:: python

            pkg_extn_details = client.package_extensions.get_details(pkg_extn_uid)
            pkg_extn_href = client.package_extensions.get_href(pkg_extn_details)

        """
        PkgExtn._validate_type(pkg_extn_details, u'pkg_extn_details', object, True)
        PkgExtn._validate_type_of_details(pkg_extn_details, PKG_EXTN_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(pkg_extn_details,
                                                           u'pkg_extn_details',
                                                           [u'entity', u'package_extension', u'href'])

    def delete(self, pkg_extn_id):
        """Delete a package extension.

        :param pkg_extn_id: Unique Id of package extension
        :type pkg_extn_id: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.package_extensions.delete(pkg_extn_id)

        """
        PkgExtn._validate_type(pkg_extn_id, u'pkg_extn_uid', str, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_id),
                                   params=self._client._params(),
                                   headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete pkg extn specification', response)

    def _delete(self, pkg_extn_uid):
        PkgExtn._validate_type(pkg_extn_uid, u'pkg_extn_uid', str, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_uid), params=self._client._params(),
                                   headers=self._client._get_headers())

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'pkg_extn_response', dict)
        try:
            if self._client.default_space_id is not None:
                new_el = {'metadata': {'space_id': response_data['metadata']['space_id'],
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:

                    href = self._client.service_instance._href_definitions.get_base_asset_href(response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

                    new_el = {'metadata': {'project_id': response_data['metadata']['project_id'],
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                           },
                              'entity': response_data['entity']

                              }
                else:
                    new_el = {'metadata': {'project_id': response_data['metadata']['project_id'],
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                       },
                             'entity': response_data['entity']

                            }
            if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and 'href' in response_data['metadata']:
                href_without_host = response_data['metadata']['href'].split('.com')[-1]
                new_el[u'metadata'].update({'href': href_without_host})

            return new_el
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data.text)

    def download(self, pkg_extn_id, filename):
        """Download a package extension.

        :param pkg_extn_id: Unique Id of the package extension to be downloaded
        :type pkg_extn_id: str

        :param filename:  filename to be used for the downloaded file
        :type filename: str

        :return: path to the downloaded package extension content
        :rtype: str

        **Example**

        .. code-block:: python

            client.package_extensions.download(pkg_extn_id,"sample_conda.yml/custom_library.zip")

        """

        PkgExtn._validate_type(pkg_extn_id, u'pkg_extn_id', str, True)

        if self._WSD:
            import urllib
            pkg_extn_response = requests.get(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_id),
                                          params=self._client._params(),
                                          headers=self._client._get_headers())

            pkg_extn_details = self._handle_response(200, u'get assets', pkg_extn_response)

            artifact_content_url = pkg_extn_details['entity']['package_extension']['href']

            r = requests.get(artifact_content_url, params=self._client._params(), headers=self._client._get_headers(),
                             stream=True)
            if r.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), r)

            downloaded_asset = r.content
            try:
                with open(filename, 'wb') as f:
                    f.write(downloaded_asset)
                print(u'Successfully saved asset content to file: \'{}\''.format(filename))
                return os.getcwd() + "/" + filename
            except IOError as e:
                raise WMLClientError(u'Saving asset with artifact_url: \'{}\' failed.'.format(filename), e)
        else:
            pkg_extn_response = requests.get(self._client.service_instance._href_definitions.get_pkg_extn_href(pkg_extn_id),
                                      params=self._client._params(),
                                      headers=self._client._get_headers())

            pkg_extn_details = self._handle_response(200, u'get assets', pkg_extn_response)

            artifact_content_url = pkg_extn_details['entity']['package_extension']['href']

            if pkg_extn_response.status_code == 200:
                if not self._ICP:
                    # att_response = requests.get(self._wml_credentials["url"]+artifact_content_url)
                    att_response = requests.get(artifact_content_url)
                else:
                    att_response = requests.get(self._wml_credentials["url"]+artifact_content_url)

                if att_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading package extension"),
                                            att_response)

                downloaded_asset = att_response.content
                try:
                    with open(filename, 'wb') as f:
                        f.write(downloaded_asset)
                    print(u'Successfully saved package extension content to file: \'{}\''.format(filename))
                    return os.getcwd() + "/" + filename
                except IOError as e:
                    raise WMLClientError(u'Saving asset with artifact_url: \'{}\' failed.'.format(filename), e)
            else:
                raise WMLClientError("Failed while downloading the package extension "+ pkg_extn_id)


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging

logger = logging.getLogger(__name__)

class Party:

    def __init__(self, **kwargs):
        from ibm_watson_machine_learning.party_wrapper import Party
        self.party = Party(**kwargs)
        self.connection = self.party.connection

    def start(self):
        self.party.start()


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function

import logging

import joblib
from sklearn.base import is_classifier
from sklearn.ensemble import HistGradientBoostingClassifier, HistGradientBoostingRegressor
from sklearn.ensemble._hist_gradient_boosting.common import PREDICTOR_RECORD_DTYPE
from sklearn.ensemble._hist_gradient_boosting.predictor import TreePredictor

from ibmfl.util.xgboost.utils import _LOSSES, is_classifier

logger = logging.getLogger(__name__)


def export_sklearn(model, full_path=None):
    """
    Auxillary function to first convert the FL XGBoost Model to Scikit Learn's
    Histogram Gradient Boosting model object, then persist the object as a
    pickled model object. This enables standalone use of the trained model
    without dependency to FL's libraries.

    Parameters/Features that are Not Supported or Transferred:
    - monotonic_cst (in v0.21)
    - warm_start
    - early_stopping
    - scoring
    - validation_fraction
    - n_iter_no_change
    - tol

    :param model: The FL XGBoost model object to convert from.
    :type  model: `XGBFLModel`
    :param full_path: The full absolute path of where the model will persist.
    :type  full_path: `str`
    :return: `None`
    """
    # Initialize Model Object (Parameter Transfer from FL Model to Sklearn)
    if model.loss == "least_squares":
        export_model = HistGradientBoostingRegressor(
            loss=model.loss,
            learning_rate=model.learning_rate,
            max_iter=model.max_iter,
            max_leaf_nodes=model.max_leaf_nodes,
            max_depth=model.max_depth,
            min_samples_leaf=model.min_samples_leaf,
            l2_regularization=model.l2_regularization,
            max_bins=model.max_bins,
            verbose=model.verbose,
            random_state=model.random_state,
        )
    elif model.loss == "categorical_crossentropy" or model.loss == "binary_crossentropy" or model.loss == "auto":
        export_model = HistGradientBoostingClassifier(
            loss=model.loss,
            learning_rate=model.learning_rate,
            max_iter=model.max_iter,
            max_leaf_nodes=model.max_leaf_nodes,
            max_depth=model.max_depth,
            min_samples_leaf=model.min_samples_leaf,
            l2_regularization=model.l2_regularization,
            max_bins=model.max_bins,
            verbose=model.verbose,
            random_state=model.random_state,
        )

    # Attribute Transfer
    export_model._baseline_prediction = model._baseline_prediction
    export_model.n_trees_per_iteration_ = model.n_trees
    export_model._n_features = model.n_features_
    export_model._bin_mapper = model.bin_mapper_

    # Model Predictor Object (Deep Object Rebuild)
    export_model._predictors = []
    for i, pred in enumerate(model._predictors):
        export_model._predictors.append([])
        for j, p in enumerate(pred):
            nodes = model._predictors[i][j].nodes.astype(PREDICTOR_RECORD_DTYPE).copy()
            export_model._predictors[i].append(TreePredictor(nodes, model.known_cat_bitsets, model.known_cat_bitsets))

    # Model Loss Function
    if hasattr(model, "classes_"):
        if model.loss == "binary_crossentropy" or len(model.classes_) == 2:
            export_model._loss = _LOSSES["binary_crossentropy"](None)
        elif model.loss == "categorical_crossentropy":
            export_model._loss = _LOSSES["categorical_crossentropy"](None)
    elif model.loss == "least_squares":
        export_model._loss = _LOSSES["least_squares"](None)

    # Target Class Encoding (For Classification Models)
    if hasattr(model, "classes_"):
        export_model.classes_ = model.classes_.ravel()

    # Perform Model Pickling
    with open(full_path, "wb") as f:
        joblib.dump(export_model, f)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function

import time
import warnings

import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ParamOutOfRange
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.foundation_models.utils.enums import ModelTypes, DecodingMethods
from ibm_watson_machine_learning.metanames import GenTextParamsMetaNames, GenTextReturnOptMetaNames
from concurrent.futures import ThreadPoolExecutor


_DEFAULT_LIST_LENGTH = 50


class Model(WMLResource):
    """Instantiate the model interface.

    :param model_id: the type of model to use
    :type model_id: str

    :param credentials: credentials to Watson Machine Learning instance
    :type credentials: dict

    :param params: parameters to use during generate requests
    :type params: dict, optional

    :param project_id: ID of the Watson Studio project
    :type project_id: str, optional

    :param space_id: ID of the Watson Studio space
    :type space_id: str, optional

    :param verify: user can pass as verify one of following:

        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    .. note::
        One of these parameters is required: ['project_id ', 'space_id']

    .. hint::
        You can copy the project_id from Project's Manage tab (Project -> Manage -> General -> Details).

    **Example**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import Model
        from ibm_watson_machine_learning.metanames import GenTextParamsMetaNames as GenParams
        from ibm_watson_machine_learning.foundation_models.utils.enums import ModelTypes, DecodingMethods

        # To display example params enter
        GenParams().get_example_values()

        generate_params = {
            GenParams.MAX_NEW_TOKENS: 25
        }

        model = Model(
            model_id=ModelTypes.FLAN_UL2,
            params=generate_params,
            credentials={
                "apikey": "***",
                "url": "https://us-south.ml.cloud.ibm.com"
            },
            project_id="*****"
            )
    """

    def __init__(self,
                 model_id: str,
                 credentials: dict,
                 params: dict = None,
                 project_id: str = None,
                 space_id: str = None,
                 verify = None) -> None:

        self.model_id = model_id
        if isinstance(self.model_id, ModelTypes):
            self.model_id = self.model_id.value
        self.params = params
        self.project_id = project_id
        self.space_id = space_id
        self._client = APIClient(credentials, verify=verify)

        Model._validate_type(params, u'params', dict, False)

        if not self._client.CLOUD_PLATFORM_SPACES and self._client.CPD_version < 4.8:
            raise WMLClientError(error_msg="Operation is unsupported for this release.")

        if space_id:
            self._client.set.default_space(space_id)
        else:
            self._client.set.default_project(project_id)

        WMLResource.__init__(self, __name__, self._client)

    def get_details(self):
        """Get model's details

        :return: model's details
        :rtype: dict

        **Example**

        .. code-block:: python

            model.get_details()

        """
        limit = 200
        asynchronous = False
        get_all = False

        url = self._client.service_instance._href_definitions.get_fm_specifications_href(limit)
        models = self._get_artifact_details(base_url=url, uid=None, limit=limit, resource_name='foundation_models', _async=asynchronous, _all=get_all)['resources']

        return next((item for item in models if item['model_id'] == self.model_id), None)

    def generate(self, prompt, params=None, concurrency_limit=10):
        """Given a text prompt as input, and parameters the selected model (model_id)
        will generate a completion text as generated_text.

        :param params: meta props for text generation, use ``ibm_watson_machine_learning.metanames.GenTextParamsMetaNames().show()`` to view the list of MetaNames
        :type params: dict

        :param concurrency_limit: number of requests that will be sent in parallel, max is 10
        :type concurrency_limit: int

        :param prompt: the prompt string or list of strings. If list of strings is passed requests will be managed in parallel with the rate of concurency_limit
        :type prompt: str, list

        :return: scoring result containing generated content
        :rtype: dict

        **Example**

        .. code-block:: python

            q = "What is 1 + 1?"
            generated_response = model.generate(prompt=q)
            print(generated_response['results'][0]['generated_text'])

        """
        generate_text_url = self._client.service_instance._href_definitions.get_fm_generation_href('text?version=2022-08-01')
        Model._validate_type(params, u'params', dict, False)
        Model._validate_type(prompt, u'prompt', [str, list], True, raise_error_for_list=True)
        Model._validate_type(concurrency_limit, 'concurrency_limit', [int, float], False, raise_error_for_list=True)

        if isinstance(concurrency_limit, float):          # convert float (ex. 10.0) to int
            concurrency_limit = int(concurrency_limit)

        if concurrency_limit > 10 or concurrency_limit < 1:
            raise ParamOutOfRange(param_name='concurrency_limit', value = concurrency_limit, min=1, max=10)

        if isinstance(prompt, list):
            generated_responses = []
            if len(prompt) <= concurrency_limit:
                with ThreadPoolExecutor() as executor:
                    response_batch = list(
                        executor.map(self.generate, prompt, [params] * len(prompt))
                    )
                generated_responses.extend(response_batch)
            else:
                for i in range(0, len(prompt), concurrency_limit):
                    prompt_batch = prompt[i:i+concurrency_limit]
                    with ThreadPoolExecutor() as executor:
                        response_batch = list(
                            executor.map(self.generate, prompt_batch, [params] * len(prompt))
                        )
                    generated_responses.extend(response_batch)
            return generated_responses

        else:
            payload = {
                "model_id": self.model_id,
                "input": prompt,
            }

            if params:
                payload['parameters'] = params
            elif self.params:
                payload['parameters'] = self.params

            if 'parameters' in payload and GenTextParamsMetaNames.DECODING_METHOD in payload['parameters']:
                if isinstance(payload['parameters'][GenTextParamsMetaNames.DECODING_METHOD], DecodingMethods):
                    payload['parameters'][GenTextParamsMetaNames.DECODING_METHOD] = payload['parameters'][GenTextParamsMetaNames.DECODING_METHOD].value

            if self._client.default_project_id:
                payload['project_id'] = self._client.default_project_id
            elif self._client.default_space_id:
                payload['space_id'] = self._client.default_space_id

            if 'parameters' in payload and 'return_options' in payload['parameters']:
                if not (payload['parameters']['return_options'].get("input_text", False) or payload['parameters']['return_options'].get("input_tokens", False)):
                    raise WMLClientError(Messages.get_message(message_id="fm_required_parameters_not_provided"))

            try:
                retries = 0
                while retries < 3:
                    response_scoring = requests.post(
                            url=generate_text_url,
                            json=payload,
                            headers=self._client._get_headers()
                        )
                    if response_scoring.status_code in [429, 503, 504, 520]:
                        time.sleep(2 ** retries)
                        retries += 1
                    else:
                        break

                if response_scoring.status_code >= 400 and response_scoring.status_code <= 599:
                    raise WMLClientError(f'Request failed with: {response_scoring.text} ({response_scoring.status_code})')

            except Exception as e:
                warnings.warn(f"Response failed with error: '{str(e)}' for {payload}")
                return {
                    'results': [{'generated_text': f"Response failed with error '{str(e)}' on prompt '{payload['input']}'", 'stop_reason': 'ERROR'}]
                }

            return self._handle_response(200, u'generate', response_scoring)

    def generate_text(self, prompt, params=None, concurrency_limit=10):
        """Given a text prompt as input, and parameters the selected model (model_id)
        will generate a completion text as generated_text.

        :param params: meta props for text generation, use ``ibm_watson_machine_learning.metanames.GenTextParamsMetaNames().show()`` to view the list of MetaNames
        :type params: dict

        :param concurrency_limit: number of requests that will be sent in parallel, max is 10
        :type concurrency_limit: int

        :param prompt: the prompt string or list of strings. If list of strings is passed requests will be managed in parallel with the rate of concurency_limit
        :type prompt: str, list

        :return: generated content
        :rtype: str

        **Example**

        .. code-block:: python

            q = "What is 1 + 1?"
            generated_text = model.generate_text(prompt=q)
            print(generated_text)

        """
        if isinstance(prompt, list):
            return [self.generate(prompt=prompt, params=params, concurrency_limit=concurrency_limit)['results'][0]['generated_text'] for prompt in prompt]
        else:
            return self.generate(prompt=prompt, params=params, concurrency_limit=concurrency_limit)['results'][0]['generated_text']



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems


class RuntimeSpecDefinitionInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, description=None, platform=None, custom_libraries=None, public_libraries=None ):
        """
        RuntimeSpecDefinitionInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'description': 'str',
            'platform': 'RuntimeSpecDefinitionInputPlatform',
            'custom_libraries': 'RuntimeSpecDefinitionInputCustomLibraries',
            'public_libraries': 'list[RuntimeSpecDefinitionInputPublicLibraries]'
        }

        self.attribute_map = {
            'name': 'name',
            'description': 'description',
            'platform': 'platform',
            'custom_libraries': 'custom_libraries',
            'public_libraries': 'public_libraries'
        }

        self._name = name
        self._description = description
        self._platform = platform
        self._custom_libraries = custom_libraries
        self._public_libraries = public_libraries

    @property
    def name(self):
        """
        Gets the name of this RuntimeSpecDefinitionInput.


        :return: The name of this RuntimeSpecDefinitionInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this RuntimeSpecDefinitionInput.


        :param name: The name of this RuntimeSpecDefinitionInput.
        :type: str
        """
        self._name = name

    @property
    def description(self):
        """
        Gets the description of this RuntimeSpecDefinitionInput.


        :return: The description of this RuntimeSpecDefinitionInput.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this RuntimeSpecDefinitionInput.


        :param description: The description of this RuntimeSpecDefinitionInput.
        :type: str
        """
        self._description = description

    @property
    def platform(self):
        """
        Gets the platform of this RuntimeSpecDefinitionInput.


        :return: The platform of this RuntimeSpecDefinitionInput.
        :rtype: RuntimeSpecDefinitionInputPlatform
        """
        return self._platform

    @platform.setter
    def platform(self, platform):
        """
        Sets the platform of this RuntimeSpecDefinitionInput.


        :param platform: The platform of this RuntimeSpecDefinitionInput.
        :type: RuntimeSpecDefinitionInputPlatform
        """
        self._platform = platform

    @property
    def custom_libraries(self):
        """
        Gets the custom_libraries of this RuntimeSpecDefinitionInput.


        :return: The custom_libraries of this RuntimeSpecDefinitionInput.
        :rtype: RuntimeSpecDefinitionInputCustomLibraries
        """
        return self._custom_libraries

    @custom_libraries.setter
    def custom_libraries(self, custom_libraries):
        """
        Sets the custom_libraries of this RuntimeSpecDefinitionInput.


        :param custom_libraries: The custom_libraries of this RuntimeSpecDefinitionInput.
        :type: RuntimeSpecDefinitionInputCustomLibraries
        """
        self._custom_libraries = custom_libraries

    @property
    def public_libraries(self):
        """
        Gets the public_libraries of this RuntimeSpecDefinitionInput.
        Array of public libraries

        :return: The public_libraries of this RuntimeSpecDefinitionInput.
        :rtype: list[RuntimeSpecDefinitionInputPublicLibraries]
        """
        return self._public_libraries

    @public_libraries.setter
    def public_libraries(self, public_libraries):
        """
        Sets the public_libraries of this RuntimeSpecDefinitionInput.
        Array of public libraries

        :param public_libraries: The public_libraries of this RuntimeSpecDefinitionInput.
        :type: list[RuntimeSpecDefinitionInputPublicLibraries]
        """
        self._public_libraries = public_libraries

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from abc import abstractmethod
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from pandas import DataFrame
    from ibm_watson_machine_learning.utils.autoai.enums import PipelineTypes
    from sklearn.pipeline import Pipeline
    from numpy import ndarray

__all__ = [
    "BaseAutoPipelines"
]


class BaseAutoPipelines:
    """Base abstract class for Pipeline Optimizers."""

    @abstractmethod
    def get_params(self) -> dict:
        """Get configuration parameters of AutoPipelines"""
        pass

    @abstractmethod
    def fit(self, *args, **kwargs) -> 'Pipeline':
        """Run fit job."""
        pass

    @abstractmethod
    def summary(self) -> 'DataFrame':
        """List all computed pipelines."""
        pass

    @abstractmethod
    def get_pipeline_details(self, pipeline_name: str = None) -> dict:
        """Get details of computed pipeline. Details like pipeline steps."""
        pass

    @abstractmethod
    def get_pipeline(self, pipeline_name: str, astype: 'PipelineTypes') -> Union['Pipeline', 'TrainablePipeline']:
        """Get particular computed Pipeline"""
        pass

    def get_pipeline_notebook(self,
                     pipeline_name: str = None,
                     persist: 'bool' = False) -> Union['Pipeline', 'TrainablePipeline']:
        """Get particular computed pipeline notebook"""
        pass

    @abstractmethod
    def predict(self, X: Union['DataFrame', 'ndarray']) -> 'ndarray':
        """Use predict on top of the computed pipeline."""
        pass



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

__all__ = [
    "LocalAutoPipelinesRuns"
]


import os

from typing import List, Dict, Union, TYPE_CHECKING, Optional

from pandas import DataFrame

from ibm_watson_machine_learning.experiment.autoai.optimizers.local_auto_pipelines import LocalAutoPipelines
from ibm_watson_machine_learning.helpers import DataConnection
from ibm_watson_machine_learning.utils.autoai.enums import DataConnectionTypes
from ibm_watson_machine_learning.utils.autoai.utils import prepare_cos_client
from ibm_watson_machine_learning.utils.autoai.watson_studio import get_project, get_wmls_credentials_and_space_ids
from ibm_watson_machine_learning.utils.autoai.training import is_run_id_exists
from ibm_watson_machine_learning.utils.autoai.errors import WrongWMLServer
from .base_auto_pipelines_runs import BaseAutoPipelinesRuns


if TYPE_CHECKING:
    from ..optimizers import RemoteAutoPipelines


class LocalAutoPipelinesRuns(BaseAutoPipelinesRuns):
    """LocalAutoPipelinesRuns class is used to work with historical Optimizer runs (local optimizer and
    with with data from COS, without WML API interaction).

    :param filter: filter, user can choose which runs to fetch specifying experiment name (option not yet available)
    :type filter: str, optional
    """

    def __init__(self, filter: str = None) -> None:
        self.experiment_name = filter
        self.training_data_reference = None
        self.training_result_reference = None

    def __call__(self, *, filter: str) -> 'LocalAutoPipelinesRuns':
        raise NotImplementedError("Not yet implemented in the local scenario.")

    def list(self) -> 'DataFrame':
        raise NotImplementedError("Not yet implemented in the local scenario.")

    def get_params(self, run_id: str = None) -> dict:
        raise NotImplementedError("Not yet implemented in the local scenario.")

    def get_run_details(self, run_id: str = None) -> dict:
        raise NotImplementedError("Not yet implemented in the local scenario.")

    def get_optimizer(self,
                      run_id: Optional[str] = None,
                      metadata: Dict[str, Union[List['DataConnection'],  'DataConnection', str, int]] = None,
                      wml_client: 'APIClient' = None
                      ) -> Union['LocalAutoPipelines', 'RemoteAutoPipelines']:
        """Get historical optimizer from historical experiment.

        :param run_id: ID of the local historical experiment run (option not yet available)
        :type run_id: str, optional

        :param metadata: option to pass information about COS data reference or WSD (auto-gen notebook)
        :type metadata: dict, optional

        :param wml_client: only for Container type to work properly
        :type wml_client: ApiClient, optional

        **Example**

        .. code-block:: python

            metadata = dict(
                   prediction_type ='classification',
                   prediction_column='species',
                   holdout_size=0.2,
                   scoring='roc_auc',
                   max_number_of_estimators=1,
                   training_data_reference = [DataConnection(
                       connection_asset_id=connection_id,
                       location=S3Location(
                           bucket='autoai-bucket',
                           path='iris_dataset.csv',
                       )
                   )],
                   training_result_reference = DataConnection(
                       connection_asset_id=connection_id,
                       location=S3Location(
                           bucket='autoai-bucket',
                           path='.',
                           model_location="0a8266be-0f3e-4ef9-af89-856022b7c1c9/data/automl/global_output/",
                           training_status="./75eec2e0-2600-4b7e-bcf2-ea54f2471400/9236e3ab-25e2-4daa-86a8-fd009d4e1e7d/training-status.json",
                       )
                   )
               )
            optimizer = AutoAI().runs.get_optimizer(metadata)
        """

        if run_id is not None:
            raise NotImplementedError("run_id option is not yet implemented in the local scenario.")

        else:
            training_result_reference = metadata.get('training_result_reference')

            # note: WSD auto-gen notebook scenario
            if training_result_reference.type == DataConnectionTypes.FS:
                run_id = training_result_reference.location.path.split('/')[-3]
                project = get_project()
                project_id = project.get_metadata()["metadata"]["guid"]

                path_to_wmls_credentials = f"{os.path.abspath('.')}/{project_id}/project.json"
                credentials, space_ids = get_wmls_credentials_and_space_ids(path_to_wmls_credentials)

                historical_optimizer = None
                for creds, space_id in zip(credentials, space_ids):

                    if is_run_id_exists(wml_credentials=creds, run_id=run_id, space_id=space_id):

                        from ibm_watson_machine_learning.experiment import AutoAI

                        historical_optimizer = AutoAI(creds, space_id=space_id).runs.get_optimizer(run_id)
                        break

                if historical_optimizer is None:
                    raise WrongWMLServer(reason="Cannot find historical AutoAI experiment in any of WML Server instances "
                                                "connected to this WML Desktop application. "
                                                "Please provide WML credentials manually.")

                else:
                    return historical_optimizer
            # --- end note

            # note: cloud auto-gen notebook scenario
            else:
                # note: save training connection to be able to further provide this data via get_data_connections
                self.training_data_reference: List['DataConnection'] = metadata.get('training_data_references',
                                                                                    metadata.get('training_data_reference'))
                self.training_result_reference: 'DataConnection' = training_result_reference

                # note: fill experiment parameters to be able to recreate holdout split
                for data in self.training_data_reference:
                    data._fill_experiment_parameters(
                        prediction_type=metadata.get('prediction_type'),
                        prediction_column=metadata.get('prediction_column'),
                        holdout_size=metadata.get('holdout_size'),
                        csv_separator=metadata.get('csv_separator', ','),
                        excel_sheet=metadata.get('excel_sheet', 0),
                        encoding=metadata.get('encoding', 'utf-8')
                    )

                self.training_result_reference._fill_experiment_parameters(
                    prediction_type=metadata.get('prediction_type'),
                    prediction_column=metadata.get('prediction_column'),
                    holdout_size=metadata.get('holdout_size'),
                    csv_separator=metadata.get('csv_separator', ','),
                    excel_sheet=metadata.get('excel_sheet', 0),
                    encoding=metadata.get('encoding', 'utf-8')
                )
                # --- end note

                # Note: We need to fetch credentials when 'container' is the type
                if (hasattr(self.training_result_reference, 'type') and
                    (self.training_result_reference.type == 'container' or
                    self.training_result_reference.type == 'data_asset' or
                     self.training_result_reference.type == 'connection_asset') and wml_client is not None):
                    self.training_result_reference.set_client(wml_client)

                for data_ref in self.training_data_reference:
                    if hasattr(data_ref, 'type') and (data_ref.type == 'container' or
                                                      data_ref.type == 'connection_asset' or
                                                      data_ref.type == 'data_asset') and wml_client is not None:
                        data_ref.set_client(wml_client)
                # --- end note

                data_clients, result_client = prepare_cos_client(
                    training_data_references=self.training_data_reference,
                    training_result_reference=self.training_result_reference
                )

                optimizer = LocalAutoPipelines(
                    name='Auto-gen notebook from COS',
                    prediction_type=metadata.get('prediction_type'),
                    prediction_column=metadata.get('prediction_column'),
                    scoring=metadata.get('scoring'),
                    holdout_size=metadata.get('holdout_size'),
                    max_num_daub_ensembles=metadata.get('max_number_of_estimators'),
                    _data_clients=data_clients,
                    _result_client=result_client
                )
                optimizer._training_data_reference = self.training_data_reference
                optimizer._training_result_reference = self.training_result_reference

                return optimizer
                # --- end note

    def get_data_connections(self, run_id: str) -> List['DataConnection']:
        raise NotImplementedError("Not yet implemented in the local scenario.")
